import java.util.ArrayList;

/**
 * Die Klasse TSP.
 * 
 * @author claus albowski
 * @version Juni 2015
 */
public class TSP
{
    private Graph graph;
    private Knoten start;
    private int anzahlKnoten;
    private Weg besterRundweg = null;

    /**
     * Konstruktor für Objekte der Klasse TSP
     */
    public TSP(int anzahlKnoten)
    {
        this.anzahlKnoten = anzahlKnoten;
        graph = new Graph(anzahlKnoten, 500,500);
        graph.vollstaendig();
        graph.zeichne();
    }

    /**
     * definiert einen vorzugebenden Knoten als Startknoten,
     * i<=anzahl !
     * und startet dann die Suche aller Rundwege
     */
    public void starteSucheMit(int i)
    {
        start = Knoten.gibKnoten(i);
        schritt(new Weg(start));
        if (besterRundweg!=null) besterRundweg.zeichneWeg();
    }

    /**
     * fuehrt den Schritt in die Tiefe aus
     */
    public void schritt(Weg weg)
    {
        //        weg.zeigeWeg();
        if (weg.istRundweg()) // neuer Rundweg gefunden
        {
            if ((besterRundweg==null)||
            (weg.wegLaenge()<besterRundweg.wegLaenge()))
            {
                besterRundweg=weg;
                besterRundweg.zeigeWeg();
            }
            return;
        }
        else if (weg.keinZyklus())
        {
            Knoten knoten = weg.gibEndKnoten();
            for (Kante kante : knoten.gibKanten())
                schritt(new Weg(weg, kante));
        }
        else return;
    }

    ///// Interne Klasse Weg /////
    private class Weg
    implements Comparable
    {        
        private ArrayList<Kante> kanten;
        private Knoten endKnoten;

        /**
         * Konstruktor fuer den Start
         */
        public Weg(Knoten startKnoten)
        {
            kanten = new ArrayList();
            endKnoten = startKnoten;
        }

        /**
         * Konstruktor fuer einen fortsetzenden Weg
         */
        public Weg(Weg weg, Kante neueKante)
        {
            kanten = new ArrayList();
            for (Kante kante : weg.gibKanten())
                kanten.add(kante);
            kanten.add(neueKante);
            endKnoten = neueKante.andererKnoten(weg.gibEndKnoten()); 
        }

        /**
         * get-Methode fuer die Kanten
         */
        public ArrayList<Kante> gibKanten() { return kanten; }

        /**
         * get-Methode fuer den Endknoten
         */
        public Knoten gibEndKnoten() { return endKnoten; }

        // Fuer das Interface  Comparable notwendige Methode:
        public int compareTo(Object obj) {
            if (wegLaenge()<((Weg)obj).wegLaenge()) return -1;
            else if (wegLaenge()>((Weg)obj).wegLaenge()) return 1;
            else return 0;
        }

        /**
         * prueft, ob die letzte eingefuegte Kante Knoten unter denen
         * vor dem alten Endknoten enthaelt
         */
        public boolean keinZyklus()
        {
            if (kanten.size()<2) return true;
            Kante letzte = kanten.get(kanten.size()-1);
            Knoten alterEndknoten = letzte.andererKnoten(endKnoten);
            if (endKnoten==start) return false;
            for (Kante kante : gibKanten())
            {
                if ((endKnoten==kante.gibP1())&&(kante!=letzte)) return false;
                if ((endKnoten==kante.gibP2())&&(kante!=letzte)) return false;
            }
            return true;
        }

        /**
         * berechnet die Weglaenge
         */
        public double wegLaenge()
        {
            double laenge = 0;
            for (Kante kante : gibKanten()) laenge += kante.laenge();
            return laenge;
        }

        /**
         * gibt zurueck, ob der Weg ein Rundweg ist
         */
        public boolean istRundweg()
        {
            return (endKnoten==start)&&(kanten.size()==anzahlKnoten);
        }

        /**
         * Anzeige in der Konsole
         */
        public void zeigeWeg()
        {
            System.out.print(start.gibNamen());
            Knoten aktuell = start;
            for (Kante kante : kanten)
            {
                aktuell = kante.andererKnoten(aktuell);
                System.out.print(" -> "+aktuell.gibNamen());
            }
            System.out.println(", Laenge "+wegLaenge());
        }

        /**
         * Anzeige im Bild
         */
        public void zeichneWeg()
        {
            Knoten aktuell = start;
            for (Kante kante : kanten)
            {
                aktuell = kante.andererKnoten(aktuell);
                kante.zeichne("rot");
            }
        }

    }
}
